'''
 Model Database (MDB) Interface

 Bjoern Annighoefer 2021
'''

from ..value import VAL, STR, NON
from ..query import Obj


class ReadOnlyMdb:
    """ Definition of an read only MDB interface.
    """
     
    def Read(self, target:Obj, featureName:STR, context:Obj=NON()) -> VAL:
        """ Returns the value of a feature.
        
        Returns the value or values of an attribute, an association or a composition. 
        The value can be a single value in case the length of the feature is 1.
        Otherwise a list of values will be returned.
        
        Args:
            target: The element the feature belongs to.
            featureName: The name of the feature of which the value shall be returned.
                Generic feature names might be used
            context: (optional) if given some feature names are only evaluated 
                below the context element, e.g. finding all class instances 
            
        Returns:
            val: The value of the feature.
            
        Raises:
            EOQ_ERROR_INVALID_VALUE: The target is no (known) element.
            EOQ_ERROR_INVALID_VALUE: The element does not have a feature with the given name 
                or it is not readable.
            
        Side-effects:
            None
        """
        
        pass




'''
 Model Database (MDB) Interface

 Bjoern Annighoefer 2021
'''

from .readonlymdb import ReadOnlyMdb

from ..value import VAL, PRM, I64, STR, LST, NON
from ..query import Obj
from ..error import EOQ_ERROR

#type checking
from typing import Tuple, List, Union


class Mdb(ReadOnlyMdb):
    """ Definition of an basic MDB interface.
    
    Defines a CRUD+ interface as being the sole interface for interacting with a model database (MDB) 
    All implementations of an MDB should inherit from this interface.
    
    * Create, read, remove, update, delete
    """
     
    def Create(self, classId:Union[STR,Obj], createArgs:LST=LST([]), target:Obj=NON(), recoveryArgs:LST=LST([])) -> Tuple[Obj,EOQ_ERROR]:
        """ Creates a new element based on a class.
        
        Instantiates a clas as a new element and returns the elements reference.
        
        Args:
            classId: The unique identifier of a class
                EOQ generic class IDs might be used as classID to create, e.g.
                meta-model elements.
            name: (optional) some classes might require providing a name
                This is usually the case for meta-model classes.
            target: (optional) if given, the new element is created as a replacement
                of the target element. This is only possible if the target element
                is from the same mdb, but not valid any more, i.e. it is an 
                element that was deleted before.                
                
        Returns:
            newElem: The element that was created, e.g. the new instance of the class.
            postModificationError: None if successful. 
                Returns instance of EOQ_ERROR, if the call failed after the 
                MDB was modified. In that case a rollback is suggested.
            
        Raises:
            EOQ_ERROR_INVALID_VALUE: The class ID provided is not known 
                and cannot be resolved in a class to be instantiated.
            EOQ_ERROR_INVALID_VALUE: The class to be instantiated requires a 
                name in order to be instantiated.
            EOQ_ERROR_INVALID_VALUE: The name given is has an invalid value or is no string 
                Or can not be used because it is already in use.
            
        Side-effects:
            None
        """
        
        pass
    
#     def Read(self, target:Obj, featureName:STR, context:Obj=NON()) -> VAL:
#         """ Returns the value of a feature.
#         
#         Returns the value or values of an attribute, an association or a composition. 
#         The value can be a single value in case the length of the feature is 1.
#         Otherwise a list of values will be returned.
#         
#         Args:
#             target: The element the feature belongs to.
#             featureName: The name of the feature of which the value shall be returned.
#                 Generic feature names might be used
#             context: (optional) if given some feature names are only evaluated 
#                 below the context element, e.g. finding all class instances 
#             
#         Returns:
#             val: The value of the feature.
#             
#         Raises:
#             EOQ_ERROR_INVALID_VALUE: The target is no (known) element.
#             EOQ_ERROR_INVALID_VALUE: The element does not have a feature with the given name 
#                 or it is not readable.
#             
#         Side-effects:
#             None
#         """
#         
#         pass
    
    def Update(self, target:Obj, featureName:STR, value:PRM, position:I64=I64(0)) -> Tuple[Obj,Obj,Obj,I64,EOQ_ERROR]:
        """ Sets or inserts a value of an element's attribute, association or composition
        
        Update is the universal tool to make changes in a model database. 
        It changes the content of a attribute, association or composition (called feature).
        
        If an value is given, this value is stored at the given position of the feature. 
        If this position already has a value, that value is replaced by the new one. 
        For compositions this will change the parent of the replaced element. 
        Likewise, it changes the parent of the new value.
        
        If None is given as the value, the current value at the given position is removed from the feature. 
        Features can not contain empty positions, therefore, the subsequent elements will 
        decrease by one in their position.
        
        The return value contains everything to undo the update. 
        Except information provided in the arguments during the call.
        
        Args:
            target: The target element of the update operation. 
                Obj(0) can be used for root operations
            featureName: The name of the feature of the target to update.
                The feature name can be a generic feature ID.
            value: value == None -> The current value at the given position is removed. 
                The subsequent elements move one position forward.
                value != None -> The current value at the given position is replaced by 
                the given value. 
            position: (optional, default = 0) Legal range [-<feature len>-2, ..., <feature len>+1]
                The position where the value shall be stored or removed. 
                Position must be positive and below the length of the feature or -1.
                If -1 is given the value is added after the last element in the feature.
                Higher negative position indicates insertion. 
                Examples: 
                    - BEFORE    --> update(... x, position) --> AFTER
                    - [1,2,3,4,5] --> upd(...,x, 0) --> [x,2,3,4,5]
                    - [1,2,3,4,5] --> upd(...,x, 2) --> [1,2,x,4,5]
                    - [1,2,3,4,5] --> upd(...,x, 5) --> [1,2,3,4,5,x]
                    - [1,2,3,4,5] --> upd(...,x, 6) --> ERROR
                    - [1,2,3,4,5] --> upd(...,x,-1) --> [1,2,3,4,5,x]
                    - [1,2,3,4,5] --> upd(...,x,-2) --> [x,1,2,3,4,5]
                    - [1,2,3,4,5] --> upd(...,x,-4) --> [1,2,x,3,4,5]
                    - [1,2,3,4,5] --> upd(...,x,-6) --> [1,2,3,4,x,5]
                    - [1,2,3,4,5] --> upd(...,x,-7) --> [1,2,3,4,5,x]
                    - [1,2,3,4,5] --> upd(...,x,-8) --> ERROR
                
        Returns:
            oldValue: The previous element or primitive value at position 
            oldOwner: The parent of value before update.
                (only in case value is an element, the feature is a composition and 
                value had a parent before, is None otherwise) 
            oldComposition: The name of the composition value was inside before the update.
                (only in case value is an element, the feature is a composition and 
                value had a parent before, is None otherwise) 
            oldPosition: The position of the element in an composition before the update.
                (only in case value is an element, the feature is a composition and 
                value had a parent before, is None otherwise) 
            postModificationError: None if successful. 
                Returns instance of EOQ_ERROR, if the call failed after the 
                MDB was modified. In that case a rollback is suggested.
            
        Raises:
            ...
            
        Side-effects:
            Parent change of old value:  (only replace of element values in compositions)
                In case of composition the parent of the old value at the position is removed
            Parent change of new value: (only if new value is an element and add to a composition)
                In case of a composition the parent of the new value is now the target
            Position change of elements > position: (insert only)
                In case of insert (negative position) the position of values after the inserted 
                value are increased by one.
        
        """
        
        pass
    
    
    def Delete(self, target:Obj) -> Tuple[STR,List[STR],List[VAL],EOQ_ERROR]: #modi for delete: complete, element only
        """ Deletes an object. 
        
        Only elements can be deleted that do not have a parent and
        that are not referenced by any other element any more.
        Any child of the deleted element does become an orphan.
        
        Args:
            target: The element to be deleted.
        
        Returns:
            classId: The classId of the deleted element 
            featureNames: All feature names of the deleted element. 
                this includes all attributes, association and composition 
                names of this and all super classes. This can be used 
                to fully restore the element if creating a new one.
            featureValues: All feature values of the deleted element
                this includes the values of all attributes, associations,
                and composition. The values are in the same order as 
                the feature names. This can be used to restore the element.
            postModificationError: None if successful. 
                Returns instance of EOQ_ERROR, if the call failed after the 
                MDB was modified. In that case a rollback is suggested.
                
        Side-effects:
            All children become orphans.
            
            Target is no valid object identifier any more, but might be used 
            for restoring the element in Create()
        """
        pass
    
    def FindElementByIdOrName(self, nameOrId:STR, context:Obj=NON(), restrictToConcept:STR=NON()) -> LST:
        ''' Returns all elements that match the given name or string id.
        
        In case an ID is provided a maximum of one class is returned, because IDs are unique.
        
        Args:
            nameOrId: The *NAME or *STRID value of an element if existent. 
                STRID will maks any name if existent, i.e. if STR ID is found, 
                no NAME is searched
            context: If set element are only used in this context. Only valid for names.
            restrictToConcept: If set, only concepts of this kind are returned, with matching ID are returned.
        
        Returns:
            classes: A list of classes matching the id or name. If no class is found, the list is empty.
        '''
        pass
    
    def Close(self):
        ''' CLoses the MDB gracefully
        
        '''
        pass




'''
 2019 Bjoern Annighoefer
'''

from ..config import Config

import queue # imported for using queue.Empty exception
from multiprocessing import Queue
from threading import Thread

import traceback



class QueueBasedFrameCom:
    def __init__(self, cmdTxQueu:Queue, cmdRxQueu:Queue,\
                config:Config,\
                rawFrmCallback):
        super().__init__()
        self.cmdTxQueu = cmdTxQueu
        self.cmdRxQueu = cmdRxQueu
        self.rawFrmCallback = rawFrmCallback
        self.config = config
        
    def SendRaw(self,frmStr:str):
        self.cmdTxQueu.put(frmStr)
        
#     def SendEvt(self,frame:Frm):
#         ##print('Sending frame on queue: %d'%(queueNb))
#         frmStr = self.config.remoteSerializer.SerFrm(frame)
#         self.SendEvtRaw(frmStr)
#         
#     def SendEvtRaw(self,frmStr:str):
#         self.evtTxQueue.put(frmStr)
#         #print('%s: event send %d bytes'%(self,len(frmStr)))
        
    def Start(self):
        ##create an event observing thread
        self.shallRun = True
        self.rxThread = Thread(target=self.__RxThread, args=('RxThread',self.cmdRxQueu,self.rawFrmCallback,))
        self.rxThread.start()
 
    def Stop(self):
        self.shallRun = False
        #print('%s: stopped (run %d)'%(self,self.shallRun))
        self.rxThread.join()
        #print('%s: Both threads stopped (run: %d)'%(self,self.shallRun))
        
    def __RxThread(self, name:str, cmdRxQueu, rawCallback):
        while(self.shallRun):
            try:
                frmStr = cmdRxQueu.get(timeout=self.config.threadLoopTimeout)
                #print("%s: frm recieved: %s"%(self,frmStr))
                try:
                    rawCallback(frmStr)
                except Exception as e:
                    print('%s: %s raw callback failed: %s'%(self,name,str(e)))
                    traceback.print_exc()
            except queue.Empty:
                pass #wait for next frame
            except Exception as e:
                print('%s: %s failed: %s'%(self,name,str(e)))
        #thread terminates here
        #print('%s: %s terminated'%(self,name))

        



'''
2019 Bjoern Annighoefer
'''

from typing import List,Callable

class LOG_LEVEL:
    DEBUG =  "debug"
    INFO =   "info"
    WARN =   "warning"
    ERROR =  "error"
    
class DEFAULT_LOGGER_LEVELS:
    L0_SILENT  = []
    L1_ERROR   = [LOG_LEVEL.ERROR]
    L2_WARNING = [LOG_LEVEL.WARN, LOG_LEVEL.ERROR]
    L3_INFO    = [LOG_LEVEL.INFO, LOG_LEVEL.WARN, LOG_LEVEL.ERROR]
    L4_DEBUG   = [LOG_LEVEL.INFO, LOG_LEVEL.WARN, LOG_LEVEL.ERROR, LOG_LEVEL.DEBUG]

class Logger:
    def __init__(self, activeLevels:List[str]=DEFAULT_LOGGER_LEVELS.L2_WARNING):
        self.activeLevels = activeLevels
        
    def ShallLog(self):
        return True
    
    def Log(self, level:str, msg:str):
        if(level in self.activeLevels and self.ShallLog()):
            self._Log(level,msg)
    
    #simple log functions
    
    def Debug(self, msg:str):
        self.Log(LOG_LEVEL.DEBUG, msg)
    
    def Info(self, msg:str):
        self.Log(LOG_LEVEL.INFO, msg)
        
    def Warn(self, msg:str):
        self.Log(LOG_LEVEL.WARN, msg)
        
    def Error(self, msg:str):
        self.Log(LOG_LEVEL.ERROR, msg)
        
    #passive log functions
    
    def PLog(self, level:str, msgFactory:Callable[[],str]):
        '''Passivatable log. Is supposed to have higher waste less
        performance if not used, because msgFactory is only evaluated if
        log level is active'''
        if(level in self.activeLevels and self.ShallLog()):
            try:
                self._Log(level,msgFactory())
            except Exception as e:
                self._Log(level,'ERROR while logging: %s'%str(e))
                
    def PDebug(self,msgFactory:Callable[[],str]):
        self.PLog(LOG_LEVEL.DEBUG, msgFactory)
        
    def PInfo(self,msgFactory:Callable[[],str]):
        self.PLog(LOG_LEVEL.INFO, msgFactory)
        
    def PWarn(self,msgFactory:Callable[[],str]):
        self.PLog(LOG_LEVEL.WARN, msgFactory)
    
    def PError(self,msgFactory:Callable[[],str]):
        self.PLog(LOG_LEVEL.ERROR, msgFactory)
    

    #the following must be overwritten to produce the output
    def _Log(self, level:str, msg:str):
        pass 

        


'''
Helper class for CmdRunner to delete elements according to different modes.
 
Bjoern Annighoefer 2022
'''

from ..concepts import M1OBJECT, M2CLASS, M2COMPOSITION
from ..value import STR, NON, LST
from .util import IsList

from typing import Tuple


class Deleter:
    def __init__(self,cmdrunner,logger):
        self.cmdrunner = cmdrunner
        self.logger = logger
        
    def DeleteAuto(self, target, tid)->Tuple[STR,LST,LST]:
        selfType = self.cmdrunner._MdbRead(tid,target,STR(M1OBJECT.CLASS))
        associates = self.cmdrunner._MdbRead(tid,target,STR(M1OBJECT.ASSOCIATES))
        #remove all associates
        for a in associates:
            self.cmdrunner._MdbUpdate(tid,a[0],a[3],NON(),a[2])
        #remove from parent 
        c = self.cmdrunner._MdbRead(tid,target,STR(M1OBJECT.PARENT))
        if(c[0]):
            featureName = self.cmdrunner._MdbRead(tid,c[1],STR(M2COMPOSITION.NAME))
            self.cmdrunner._MdbUpdate(tid,c[0],featureName,NON(),c[2])
        #now element can be deleted from mdb
        return self.cmdrunner._MdbDelete(tid,target)
        
        
    def DeleteFull(self, target, tid)->Tuple[STR,LST,LST]:
        clazz = self.cmdrunner._MdbRead(tid,target,STR(M1OBJECT.CLASS))
        compositions = self.cmdrunner._MdbRead(tid,clazz,STR(M2CLASS.ALLSRCCOMPOSITIONS))
        for c in compositions:
            cname = self.cmdrunner._MdbRead(tid,c,STR(M2COMPOSITION.NAME))
            children = self.cmdrunner._MdbRead(tid,target,cname)
            if(IsList(children)):
                #remove children from parent
                nChildren = len(children)
                for i in range(nChildren-1,-1,-1):
                    child = children[i]
                    self.DeleteFull(child,tid)
            elif(not children.IsNone()): 
                self.DeleteFull(children,tid)
        return self.DeleteAuto(target,tid)


from .cmdstream import CmdStreamA, CmdStream, RecorderCmdStream, ResNameGenStream, DependencyResolverStream, ObjToHisStream, ToDomainCmdStream, FromDomainCmdStream, EoqFileInStream, EoqFileOutStream
from .resnamegen import SimpleResNameGen, FromDomainResNameGen
from .validateeoqfile import ValidateEoqFile
from .loadeoqfile import LoadEoqFile
from .saveeoqfile import SaveEoqFile


'''
 conceptwalker.py a class that can walk through the clone path of concepts. 
 This is meant to be inherited by other classes to carry out certain actions.
 
 This file is generated form: conceptwalker.py.mako on 2024-07-10 10:00:02.038862 by calling E:/1_ILS/Projects/2018_EOQ/Groups/PYEOQ/pyeoq/eoq3conceptsgen/generatefromconceptscli.py -i conceptwalker.py.mako -o ../conceptwalker.py.
 
 Bjoern Annighoefer 2023
'''

from ..concepts import CONCEPTS, MXMDB ,MXELEMENT ,MXCONSTRAINT ,M2PRIMITIVES ,M2PACKAGE ,M2MODEL ,M2ENUM ,M2OPTIONOFENUM ,M2CLASS ,M2ATTRIBUTE ,M2ASSOCIATION ,M2COMPOSITION ,M2INHERITANCE ,M1MODEL ,M1OBJECT ,M1FEATURE ,M1ATTRIBUTE ,M1ASSOCIATION ,M1COMPOSITION
from ..value import STR, NON, LST
from ..query import Obj
from ..error import EOQ_ERROR_RUNTIME

from typing import Any


class ConceptWalker:
    def __init__(self, cmdrunner):
        self.cmdrunner = cmdrunner
        self.walkerHandlers = {}
        self.walkerHandlers["*MXMDB"] = self._WalkMxMdb
        self.walkerHandlers["*MXCONSTRAINT"] = self._WalkMxConstraint
        self.walkerHandlers["*M2PACKAGE"] = self._WalkM2Package
        self.walkerHandlers["*M2MODEL"] = self._WalkM2Model
        self.walkerHandlers["*M2ENUM"] = self._WalkM2Enum
        self.walkerHandlers["*M2OPTIONOFENUM"] = self._WalkM2OptionOfEnum
        self.walkerHandlers["*M2CLASS"] = self._WalkM2Class
        self.walkerHandlers["*M2ATTRIBUTE"] = self._WalkM2Attribute
        self.walkerHandlers["*M2ASSOCIATION"] = self._WalkM2Association
        self.walkerHandlers["*M2COMPOSITION"] = self._WalkM2Composition
        self.walkerHandlers["*M2INHERITANCE"] = self._WalkM2Inheritance
        self.walkerHandlers["*M1MODEL"] = self._WalkM1Model
        self.walkerHandlers["*M1OBJECT"] = self._WalkM1Object
        self.walkerHandlers["*M1ATTRIBUTE"] = self._WalkM1Attribute
        self.walkerHandlers["*M1ASSOCIATION"] = self._WalkM1Association
        self.walkerHandlers["*M1COMPOSITION"] = self._WalkM1Composition

    def Walk(self, tid:int, target:Obj, successor:Obj, maxRecursion:int=-1, data:Any=None)->Any:
        conceptId = self.cmdrunner._MdbRead(tid,target,STR(MXELEMENT.CONCEPT)).GetVal()
        try:
            handler = self.walkerHandlers[conceptId]
            res = handler(tid,target,successor,maxRecursion,data)
        except KeyError:
            raise EOQ_ERROR_RUNTIME("Cannot walk element %s of type %s"%(str(target),conceptId))

    def _WalkMxMdb(self, tid:int, target:Obj, successor:Obj, maxRecursion:int, data:Any)->Any:
        self._OnEnterMxMdb(tid,target,successor,maxRecursion,data)
        self._OnExitMxMdb(tid,target,successor,maxRecursion,data)

    def _WalkMxElement(self, tid:int, target:Obj, successor:Obj, maxRecursion:int, data:Any)->Any:
        self._OnEnterMxElement(tid,target,successor,maxRecursion,data)
        if(maxRecursion!=0):
            # walk clone path properties
            # CONSTRAINTS
            constraints = self.cmdrunner._MdbRead(tid,target,STR(MXELEMENT.CONSTRAINTS))
            for e in constraints:
                self._WalkMxConstraint(tid,e,target,maxRecursion-1,data)
        self._OnExitMxElement(tid,target,successor,maxRecursion,data)

    def _WalkMxConstraint(self, tid:int, target:Obj, successor:Obj, maxRecursion:int, data:Any)->Any:
        self._OnEnterMxConstraint(tid,target,successor,maxRecursion,data)
        # walk super concepts
        self._WalkMxElement(tid,target,target,maxRecursion,data)
        self._OnExitMxConstraint(tid,target,successor,maxRecursion,data)

    def _WalkM2Primitives(self, tid:int, target:Obj, successor:Obj, maxRecursion:int, data:Any)->Any:
        self._OnEnterM2Primitives(tid,target,successor,maxRecursion,data)
        self._OnExitM2Primitives(tid,target,successor,maxRecursion,data)

    def _WalkM2Package(self, tid:int, target:Obj, successor:Obj, maxRecursion:int, data:Any)->Any:
        self._OnEnterM2Package(tid,target,successor,maxRecursion,data)
        # walk super concepts
        self._WalkMxElement(tid,target,target,maxRecursion,data)
        if(maxRecursion!=0):
            # walk clone path properties
            # SUBPACKAGES
            subpackages = self.cmdrunner._MdbRead(tid,target,STR(M2PACKAGE.SUBPACKAGES))
            for e in subpackages:
                self._WalkM2Package(tid,e,target,maxRecursion-1,data)
            # CLASSES
            classes = self.cmdrunner._MdbRead(tid,target,STR(M2PACKAGE.CLASSES))
            for e in classes:
                self._WalkM2Class(tid,e,target,maxRecursion-1,data)
            # ENUMS
            enums = self.cmdrunner._MdbRead(tid,target,STR(M2PACKAGE.ENUMS))
            for e in enums:
                self._WalkM2Enum(tid,e,target,maxRecursion-1,data)
        self._OnExitM2Package(tid,target,successor,maxRecursion,data)

    def _WalkM2Model(self, tid:int, target:Obj, successor:Obj, maxRecursion:int, data:Any)->Any:
        self._OnEnterM2Model(tid,target,successor,maxRecursion,data)
        # walk super concepts
        self._WalkM2Package(tid,target,target,maxRecursion,data)
        self._OnExitM2Model(tid,target,successor,maxRecursion,data)

    def _WalkM2Enum(self, tid:int, target:Obj, successor:Obj, maxRecursion:int, data:Any)->Any:
        self._OnEnterM2Enum(tid,target,successor,maxRecursion,data)
        # walk super concepts
        self._WalkMxElement(tid,target,target,maxRecursion,data)
        if(maxRecursion!=0):
            # walk clone path properties
            # OPTIONS
            options = self.cmdrunner._MdbRead(tid,target,STR(M2ENUM.OPTIONS))
            for e in options:
                self._WalkM2OptionOfEnum(tid,e,target,maxRecursion-1,data)
        self._OnExitM2Enum(tid,target,successor,maxRecursion,data)

    def _WalkM2OptionOfEnum(self, tid:int, target:Obj, successor:Obj, maxRecursion:int, data:Any)->Any:
        self._OnEnterM2OptionOfEnum(tid,target,successor,maxRecursion,data)
        # walk super concepts
        self._WalkMxElement(tid,target,target,maxRecursion,data)
        self._OnExitM2OptionOfEnum(tid,target,successor,maxRecursion,data)

    def _WalkM2Class(self, tid:int, target:Obj, successor:Obj, maxRecursion:int, data:Any)->Any:
        self._OnEnterM2Class(tid,target,successor,maxRecursion,data)
        # walk super concepts
        self._WalkMxElement(tid,target,target,maxRecursion,data)
        if(maxRecursion!=0):
            # walk clone path properties
            # MYATTRIBUTES
            myattributes = self.cmdrunner._MdbRead(tid,target,STR(M2CLASS.MYATTRIBUTES))
            for e in myattributes:
                self._WalkM2Attribute(tid,e,target,maxRecursion-1,data)
            # MYSRCASSOCIATIONS
            mysrcassociations = self.cmdrunner._MdbRead(tid,target,STR(M2CLASS.MYSRCASSOCIATIONS))
            for e in mysrcassociations:
                self._WalkM2Association(tid,e,target,maxRecursion-1,data)
            # MYPARENTCOMPOSITIONS
            myparentcompositions = self.cmdrunner._MdbRead(tid,target,STR(M2CLASS.MYPARENTCOMPOSITIONS))
            for e in myparentcompositions:
                self._WalkM2Composition(tid,e,target,maxRecursion-1,data)
            # MYGENERALIZATIONS
            mygeneralizations = self.cmdrunner._MdbRead(tid,target,STR(M2CLASS.MYGENERALIZATIONS))
            for e in mygeneralizations:
                self._WalkM2Inheritance(tid,e,target,maxRecursion-1,data)
        self._OnExitM2Class(tid,target,successor,maxRecursion,data)

    def _WalkM2Attribute(self, tid:int, target:Obj, successor:Obj, maxRecursion:int, data:Any)->Any:
        self._OnEnterM2Attribute(tid,target,successor,maxRecursion,data)
        # walk super concepts
        self._WalkMxElement(tid,target,target,maxRecursion,data)
        self._OnExitM2Attribute(tid,target,successor,maxRecursion,data)

    def _WalkM2Association(self, tid:int, target:Obj, successor:Obj, maxRecursion:int, data:Any)->Any:
        self._OnEnterM2Association(tid,target,successor,maxRecursion,data)
        # walk super concepts
        self._WalkMxElement(tid,target,target,maxRecursion,data)
        self._OnExitM2Association(tid,target,successor,maxRecursion,data)

    def _WalkM2Composition(self, tid:int, target:Obj, successor:Obj, maxRecursion:int, data:Any)->Any:
        self._OnEnterM2Composition(tid,target,successor,maxRecursion,data)
        # walk super concepts
        self._WalkMxElement(tid,target,target,maxRecursion,data)
        self._OnExitM2Composition(tid,target,successor,maxRecursion,data)

    def _WalkM2Inheritance(self, tid:int, target:Obj, successor:Obj, maxRecursion:int, data:Any)->Any:
        self._OnEnterM2Inheritance(tid,target,successor,maxRecursion,data)
        # walk super concepts
        self._WalkMxElement(tid,target,target,maxRecursion,data)
        self._OnExitM2Inheritance(tid,target,successor,maxRecursion,data)

    def _WalkM1Model(self, tid:int, target:Obj, successor:Obj, maxRecursion:int, data:Any)->Any:
        self._OnEnterM1Model(tid,target,successor,maxRecursion,data)
        # walk super concepts
        self._WalkMxElement(tid,target,target,maxRecursion,data)
        if(maxRecursion!=0):
            # walk clone path properties
            # OBJECTS
            objects = self.cmdrunner._MdbRead(tid,target,STR(M1MODEL.OBJECTS))
            for e in objects:
                self._WalkM1Object(tid,e,target,maxRecursion-1,data)
        self._OnExitM1Model(tid,target,successor,maxRecursion,data)

    def _WalkM1Object(self, tid:int, target:Obj, successor:Obj, maxRecursion:int, data:Any)->Any:
        self._OnEnterM1Object(tid,target,successor,maxRecursion,data)
        # walk super concepts
        self._WalkMxElement(tid,target,target,maxRecursion,data)
        if(maxRecursion!=0):
            # walk clone path properties
            # ATTRIBUTES
            attributes = self.cmdrunner._MdbRead(tid,target,STR(M1OBJECT.ATTRIBUTES))
            for e in attributes:
                self._WalkM1Attribute(tid,e,target,maxRecursion-1,data)
            # SRCASSOCIATIONS
            srcassociations = self.cmdrunner._MdbRead(tid,target,STR(M1OBJECT.SRCASSOCIATIONS))
            for e in srcassociations:
                self._WalkM1Association(tid,e,target,maxRecursion-1,data)
            # PARENTCOMPOSITIONS
            parentcompositions = self.cmdrunner._MdbRead(tid,target,STR(M1OBJECT.PARENTCOMPOSITIONS))
            for e in parentcompositions:
                self._WalkM1Composition(tid,e,target,maxRecursion-1,data)
        self._OnExitM1Object(tid,target,successor,maxRecursion,data)

    def _WalkM1Feature(self, tid:int, target:Obj, successor:Obj, maxRecursion:int, data:Any)->Any:
        self._OnEnterM1Feature(tid,target,successor,maxRecursion,data)
        # walk super concepts
        self._WalkMxElement(tid,target,target,maxRecursion,data)
        self._OnExitM1Feature(tid,target,successor,maxRecursion,data)

    def _WalkM1Attribute(self, tid:int, target:Obj, successor:Obj, maxRecursion:int, data:Any)->Any:
        self._OnEnterM1Attribute(tid,target,successor,maxRecursion,data)
        # walk super concepts
        self._WalkM1Feature(tid,target,target,maxRecursion,data)
        self._OnExitM1Attribute(tid,target,successor,maxRecursion,data)

    def _WalkM1Association(self, tid:int, target:Obj, successor:Obj, maxRecursion:int, data:Any)->Any:
        self._OnEnterM1Association(tid,target,successor,maxRecursion,data)
        # walk super concepts
        self._WalkM1Feature(tid,target,target,maxRecursion,data)
        self._OnExitM1Association(tid,target,successor,maxRecursion,data)

    def _WalkM1Composition(self, tid:int, target:Obj, successor:Obj, maxRecursion:int, data:Any)->Any:
        self._OnEnterM1Composition(tid,target,successor,maxRecursion,data)
        # walk super concepts
        self._WalkM1Feature(tid,target,target,maxRecursion,data)
        if(maxRecursion!=0):
            # walk clone path properties
            # CHILD
            e = self.cmdrunner._MdbRead(tid,target,STR(M1COMPOSITION.CHILD))
            if not e.IsNone(): self.Walk(tid,e,target,maxRecursion-1,data)
        self._OnExitM1Composition(tid,target,successor,maxRecursion,data)


    def _OnEnterMxMdb(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnEnterMxElement(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnEnterMxConstraint(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnExitMxMdb(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnExitMxElement(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnExitMxConstraint(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnEnterM2Primitives(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnEnterM2Package(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnEnterM2Model(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnEnterM2Enum(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnEnterM2OptionOfEnum(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnEnterM2Class(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnEnterM2Attribute(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnEnterM2Association(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnEnterM2Composition(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnEnterM2Inheritance(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnExitM2Primitives(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnExitM2Package(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnExitM2Model(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnExitM2Enum(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnExitM2OptionOfEnum(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnExitM2Class(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnExitM2Attribute(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnExitM2Association(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnExitM2Composition(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnExitM2Inheritance(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnEnterM1Model(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnEnterM1Object(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnEnterM1Feature(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnEnterM1Attribute(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnEnterM1Association(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnEnterM1Composition(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnExitM1Model(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnExitM1Object(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnExitM1Feature(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnExitM1Attribute(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnExitM1Association(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior

    def _OnExitM1Composition(self, tid:int, target:Obj, successor:Obj, recursionLevel:int, data:Any)->Any:
        pass #override to implement dedicated behavior




from .logger import Logger,\
    LOG_LEVEL,\
    DEFAULT_LOGGER_LEVELS
from .nologger import NoLogger
from .consolelogger import ConsoleLogger
from .consoleandfilelogger import ConsoleAndFileLogger
from .filelogger import FileLogger



from .error import\
    EOQ_ERROR,\
    EOQ_ERROR_CODES,\
    EOQ_ERROR_INVALID_TYPE,\
    EOQ_ERROR_INVALID_VALUE,\
    EOQ_ERROR_INVALID_OPERATION,\
    EOQ_ERROR_DOES_NOT_EXIST,\
    EOQ_ERROR_RUNTIME,\
    EOQ_ERROR_ACCESS_DENIED,\
    EOQ_ERROR_UNSUPPORTED,\
    EOQ_ERROR_UNKNOWN,\
    EoqErrorFactory,\
    EoqErrorAsString



from .accesscontroller import AccessController
from .noaccesscontroller import NoAccessController
from .permissions import\
     PERMISSION_FLAGS,\
     PermissionToStr, StrToPermission,\
     PERMISSION_STR_SEPERATOR, WILDCARD_FEATURE_NAME,\
     GenerateFeaturePermissionStr,\
     ParseFeaturePermissionStr



'''
2019 Bjoern Annighoefer
'''

from .logger import Logger,DEFAULT_LOGGER_LEVELS

import os
import logging

class ConsoleAndFileLogger(Logger):
    '''A logger that outputs every thing to the console and dedicated files for each active log level
    '''
    def __init__(self,toConsole=True,toFile=True,logDir='./log',activeLevels=DEFAULT_LOGGER_LEVELS.L2_WARNING):
        super().__init__(activeLevels)
        self.toConsole = toConsole
        self.toFile = toFile
        self.logDir = logDir
        
        self.loggers = {}
        
        if(self.toFile):
            #make sure the
            if(not os.path.isdir(self.logDir)):
                os.makedirs(self.logDir)
            #create native python loggers for each level
            for level in self.activeLevels:
                #init error logger
                logger = logging.getLogger(level)
                logFile = os.path.join(self.logDir,"%s.log"%(level))
                fh = logging.FileHandler(logFile,'w')
                fh.setLevel(logging.INFO)
                formatter = logging.Formatter('%(asctime)s - %(message)s')
                fh.setFormatter(formatter)
                logger.addHandler(fh)
                logger.setLevel(logging.INFO)
                self.loggers[level] = logger
                
    #@Override
    def ShallLog(self):
        return (self.toConsole or self.toFile)
                
    #@Override         
    def _Log(self,level,msg):
        if(self.toConsole):
            print("%s: %s"%(level,msg))
        if(self.toFile):
            self.loggers[level].info(msg)


'''
Helper class for CmdRunner to clone elements according to different modes.
 
Bjoern Annighoefer 2022
'''


from ..command import CLO_MODES
from ..concepts import M1OBJECT, M2CLASS, M2ATTRIBUTE, M2ASSOCIATION, M2COMPOSITION
from ..value import STR,I64,NON
from .util import IsList

class Cloner:
    def __init__(self,cmdrunner,logger):
        self.cmdrunner = cmdrunner
        self.logger = logger
        
    def Clone(self, obj, mode, tid):
        clone = NON()
        if(CLO_MODES.CLS==mode):
            (clone,clazz) = self.__ECloneClass(obj,tid)
        elif(CLO_MODES.ATT==mode):
            (clone,clazz) = self.__ECloneAttributes(obj,tid)
        elif(CLO_MODES.FLT==mode):
            (clone,clazz) = self.__ECloneFlat(obj,tid)
        elif(CLO_MODES.FUL==mode):
            (clone,clazz) = self.__ECloneFull(obj,tid)
        elif(CLO_MODES.DEP==mode):
            (clone,clazz) = self.__ECloneDeep(obj,tid)
        return clone
        
    def __ECloneClass(self,obj,tid):
        clazz = self.cmdrunner._MdbRead(tid,obj,STR(M1OBJECT.CLASS))
        #classId = self.cmdrunner._MdbRead(tid,clazz,STR(M2CLASS.ID))
        clone = self.cmdrunner._MdbCreate(tid,clazz)
        return (clone,clazz)
    
    def __ECloneAttributes(self,obj,tid):
        (clone,clazz) = self.__ECloneClass(obj,tid)
        #attributes
        attributes = self.cmdrunner._MdbRead(tid,clazz,STR(M2CLASS.ALLATTRIBUTES))
        for attribute in attributes:
            featureName = self.cmdrunner._MdbRead(tid,attribute,STR(M2ATTRIBUTE.NAME))
            value = self.cmdrunner._MdbRead(tid,obj,featureName)
            if(IsList(value)):
                for v in value:
                    self.cmdrunner._MdbUpdate(tid,clone,featureName,v,I64(-1))
            elif(value): #do not copy empty single value features
                self.cmdrunner._MdbUpdate(tid,clone,featureName,value)
        return (clone,clazz)
    
    def __ECloneFlat(self,obj,tid):
        (clone,clazz) = self.__ECloneAttributes(obj,tid)
        associations = self.cmdrunner._MdbRead(tid,clazz,STR(M2CLASS.ALLSRCASSOCIATIONS))
        for association in associations:
            srcType = self.cmdrunner._MdbRead(tid,association,STR(M2ASSOCIATION.SRCCLASS))
            if(srcType == clazz): #prevent cloning references in the opposite direction
                featureName = self.cmdrunner._MdbRead(tid,association,STR(M2ASSOCIATION.DSTNAME))
                value = self.cmdrunner._MdbRead(tid,obj,featureName)
                if(IsList(value)):
                    for v in value:
                        self.cmdrunner._MdbUpdate(tid,clone,featureName,v,I64(-1))
                elif(value): #do not copy empty single value features
                    self.cmdrunner._MdbUpdate(tid,clone,featureName,value)
        return (clone,clazz)
    
    def __ECloneDeep(self,obj,tid,copyReferences=True,clonedElementLut={}):
        (clone,clazz) = self.__ECloneAttributes(obj,tid)
        clonedElementLut[obj.v[0].v[0]] = clone #remember what elements have been cloned
        compositions = self.cmdrunner._MdbRead(tid,clazz,STR(M2CLASS.ALLSRCCOMPOSITIONS))
        for composition in compositions:
            featureName = self.cmdrunner._MdbRead(tid,composition,STR(M2COMPOSITION.NAME))
            children = self.cmdrunner._MdbRead(tid,obj,featureName)
            if(IsList(children)):
                for child in children:
                    (clonedChild,childClass) = self.__ECloneDeep(child,tid,copyReferences,clonedElementLut)
                    self.cmdrunner._MdbUpdate(tid,clone,featureName,clonedChild,I64(-1))
            elif(children): #do not copy empty single value features
                (clonedChild,childClass) = self.__ECloneDeep(children,tid,copyReferences,clonedElementLut)
                self.cmdrunner._MdbUpdate(tid,clone,featureName,clonedChild)
        if(copyReferences): #non containments
            associations = self.cmdrunner._MdbRead(tid,clazz,STR(M2CLASS.ALLSRCASSOCIATIONS))
            for association in associations:
                featureName = self.cmdrunner._MdbRead(tid,association,STR(M2ASSOCIATION.DSTNAME))
                value = self.cmdrunner._MdbRead(tid,obj,featureName)
                if(IsList(value)):
                    for v in value:
                        self.cmdrunner._MdbUpdate(tid,clone,featureName,v,I64(-1))
                elif(value): #do not copy empty single value features
                    self.cmdrunner._MdbUpdate(tid,clone,featureName,value)
            
        return (clone,clazz)
    
    def __ECloneFull(self,obj,tid):
        clonedElementLut = {}
        (clone,clazz) = self.__ECloneDeep(obj,tid,False,clonedElementLut)
        self.__ECloneFullReferenceUpdater(obj,tid,clonedElementLut)            
        return (clone,clazz)
    
    def __ECloneFullReferenceUpdater(self,obj,tid,clonedElementLut):
        clazz = self.cmdrunner._MdbRead(tid,obj,STR(M1OBJECT.CLASS))
        #iterate through all children
        compositions = self.cmdrunner._MdbRead(tid,clazz,STR(M2CLASS.ALLSRCCOMPOSITIONS))
        for composition in compositions:
            featureName = self.cmdrunner._MdbRead(tid,composition,STR(M2COMPOSITION.NAME))
            children = self.cmdrunner._MdbRead(tid,obj,featureName)
            if(IsList(children)):
                for child in children:
                    self.__ECloneFullReferenceUpdater(child,tid,clonedElementLut)
            elif(children): #do not copy empty single value features
                self.__ECloneFullReferenceUpdater(children,tid,clonedElementLut)
        #set the references accordingly
        clone = clonedElementLut[obj.v[0].v[0]]
        associations = self.cmdrunner._MdbRead(tid,clazz,STR(M2CLASS.ALLSRCASSOCIATIONS))
        for association in associations:
            featureName = self.cmdrunner._MdbRead(tid,association,STR(M2ASSOCIATION.DSTNAME))
            value = self.cmdrunner._MdbRead(tid,obj,featureName)
            if(IsList(value)):
                for v in value:
                    ref = self.__ECloneFullFindCorrespondingReference(v, clonedElementLut)
                    self.cmdrunner._MdbUpdate(tid,clone,featureName,ref,I64(-1))
            elif(value): #do not copy empty single value features
                ref = self.__ECloneFullFindCorrespondingReference(value, clonedElementLut)
                self.cmdrunner._MdbUpdate(tid,clone,featureName,ref)
        

    
    def __ECloneFullFindCorrespondingReference(self,obj,clonedElementLut):
        if(obj.v[0].v[0] in clonedElementLut):
            return clonedElementLut[obj.v[0].v[0]]
        else:
            return obj



'''
 concepts.py provides a generic interface each model back-end working with eoq3 should support. 
 Concepts are generic constructs that should be available in any domain-specific framework and 
 modeling language. 
 
 This file is generated form: gen/concepts.py.mako on 2024-07-11 15:05:08.023707 by calling E:/1_ILS/Projects/2018_EOQ/Groups/PYEOQ/pyeoq/eoq3conceptsgen/generatefromconceptscli.py -i gen/concepts.py.mako -o concepts.py.
 
 Bjoern Annighoefer 2023
'''

'''
DEFINITIONS 
'''

CONCEPT_PREFIX = '*'
FEATURE_MULTIVALUE_POSTFIX = '*'
CONCEPT_PREFIX_LEN = len(CONCEPT_PREFIX)
CONCEPT_UNIQUE_LEN = 4 #statically defined. Checked in generate concepts


class CONCEPTS:
    def __init__(self):
        raise NotImplementedError()
    MXMDB = "*MXMDB" # Create the MDB to get the ID. Always the same ID is returned.
    MXELEMENT = "*MXELEMENT" # 
    MXCONSTRAINT = "*MXCONSTRAINT" # 
    M2PRIMITIVES = "*M2PRIMITIVES" # The primitive data types. This is used to type M2 attributes. This concept cannot be instantiated.
    M2PACKAGE = "*M2PACKAGE" # A package is the container of M2 elements such as classes an enums. 
    M2MODEL = "*M2MODEL" # An M2 model is the root package of a M2 model.
    M2ENUM = "*M2ENUM" # An enumeration data type.
    M2OPTIONOFENUM = "*M2OPTIONOFENUM" # One option of one enumeration data type.
    M2CLASS = "*M2CLASS" # A class definition.
    M2ATTRIBUTE = "*M2ATTRIBUTE" # The definition of an attribute feature for an M2 class.
    M2ASSOCIATION = "*M2ASSOCIATION" # The definition of an association feature for an M2 class.
    M2COMPOSITION = "*M2COMPOSITION" # The definition of a composition feature for an M2 class.
    M2INHERITANCE = "*M2INHERITANCE" # An inheritance defines a bilateral inheritance of one M2 class from another.
    M1MODEL = "*M1MODEL" # A container for one user mode.
    M1OBJECT = "*M1OBJECT" # An instance of an M2 class.
    M1FEATURE = "*M1FEATURE" # 
    M1ATTRIBUTE = "*M1ATTRIBUTE" # An instance of an M2 attribute for one M1 object and one value.
    M1ASSOCIATION = "*M1ASSOCIATION" # An instance of an M2 association for a relation of two M1 objects.
    M1COMPOSITION = "*M1COMPOSITION" # An instance of an M2 composition for the containment of one M1 object in on other M1 object.


'''
Mx LAYER
Elements that can exist in M1 or M2 layer or outside of them.
'''
class MXMDB:
    def __init__(self):
        raise NotImplementedError()
    CONCEPT = "*CONCEPT" #   STR   1   1 The concept ID of the concept.
    ORPHANS = "*ORPHANS" # MxElement   0   1 Elements without a parent. WARN: NOT PORTABLE
    M2MODELS = "*M2MODELS*" # M2Model   0  -1 All M2 models
    M1MODELS = "*M1MODELS*" # M1Model   0  -1 All M1 models
    ROOT = "*ROOT" # MxElement   0   1 The root element of the MDB
    MDB = "*MDB" # MxMdb   1   1 The MXMDB of this element. In this case it is pointing to myself
    
class MXELEMENT:
    def __init__(self):
        raise NotImplementedError()
    CONCEPT = "*CONCEPT" #   STR   1   1 The concept ID string
    STRID = "*STRID" #   STR   0   1 A unique ID within the MDB
    DOCUMENTATION = "*DOCUMENTATION" #   STR   0   1 A string describing this model element
    OWNER = "*OWNER" #   STR   0   1 TBD: The user name of the owner like in Unix file systems
    GROUP = "*GROUP" #   STR   0   1 TBD: The group name of the owing group like in Unix file systems
    PERMISSIONS = "*PERMISSIONS*" #   STR   0  -1 TBD: A list of string values where each string specifies the read write and create rights for the element or subelements
    HASH = "*HASH" #   I64   0   1 TBD: HASH shall make elements addressable independent of internally assigned IDs, but this is only an idea so far.
    CONSTRAINTS = "*CONSTRAINTS*" # MxConstraint   0  -1 All constraints attached to this element
    MDB = "*MDB" # MxMdb   1   1 The MXMDB of this element. In this case it is pointing to myself
    
class MXCONSTRAINT:
    def __init__(self):
        raise NotImplementedError()
    ELEMENT = "*ELEMENT" # MxElement   1   1 The element this constraint is attached to
    EXPRESSION = "*EXPRESSION" #   STR   1   1 A string holding the constraint expression. This must be a valid query resulting in a boolean.
    
'''
M2 LAYER
The meta-model layer.
'''
class M2PRIMITIVES:
    def __init__(self):
        raise NotImplementedError()
    BOL = "*BOL" #   BOL   0   0 Boolean value
    U32 = "*U32" #   U32   0   0 Unsigned integer 32 bit value
    U64 = "*U64" #   U64   0   0 Unsigned integer 64 bit value
    I32 = "*I32" #   I32   0   0 Signed integer 32 bit value
    I64 = "*I64" #   I64   0   0 Signed integer 64 bit value
    F32 = "*F32" #   F32   0   0 Floating point 32 bit value
    F64 = "*F64" #   F64   0   0 Floating point 64 bit value
    STR = "*STR" #   STR   0   0 String value
    DAT = "*DAT" #   DAT   0   0 Date and time value
    ENU = "*ENU" #   ENU   0   0 Enum value. This is only a place holder for enums defined in the meta model.
    
class M2PACKAGE:
    def __init__(self):
        raise NotImplementedError()
    NAME = "*NAME" #   STR   1   1 Name of the package
    SUPERPACKAGE = "*SUPERPACKAGE" # M2Package   0   1 The parent package or m2 model this package is contained in.
    SUBPACKAGES = "*SUBPACKAGES*" # M2Package   0  -1 All contained subpackage
    CLASSES = "*CLASSES*" # M2Class   0  -1 All M2 classes
    ENUMS = "*ENUMS*" # M2Enum   0  -1 All M2 Enums
    M1MODELS = "*M1MODELS*" # M1Model   0  -1 All M1 models instantiating this package
    
class M2MODEL:
    def __init__(self):
        raise NotImplementedError()
    
class M2ENUM:
    def __init__(self):
        raise NotImplementedError()
    NAME = "*NAME" #   STR   1   1 The name of the enum
    PACKAGE = "*PACKAGE" # M2Package   1   1 The package the enum is contained in.
    OPTIONS = "*OPTIONS*" # M2OptionOfEnum   0  -1 All options of the enum.
    ATTRIBUTES = "*ATTRIBUTES*" # M2Attribute   0  -1 All M2 attributes using this enum as type
    
class M2OPTIONOFENUM:
    def __init__(self):
        raise NotImplementedError()
    NAME = "*NAME" #   STR   1   1 The name of the enum option. This is used for setting the enum
    VALUE = "*VALUE" #   U64   1   1 An optional integer value for the enum option
    ENUM = "*ENUM" # M2Enum   1   1 The enum this option belongs to.
    M1ATTRIBUTESUSINGOPTION = "*M1ATTRIBUTESUSINGOPTION*" # M1Attribute   0  -1 All instances of attributes on M1 level, that use this value
    
class M2CLASS:
    def __init__(self):
        raise NotImplementedError()
    NAME = "*NAME" #   STR   1   1 The name of the M2 class
    ISABSTRACT = "*ISABSTRACT" #   BOL   1   1 Whether the class is abstract or not.
    PACKAGE = "*PACKAGE" # M2Package   1   1 Whether this class is abstract or not.
    INSTANCES = "*INSTANCES*" # M1Object   0  -1 All instances of this or derived classes. This is not a delete blocker, because if inherit classes exist the inheritance will prevent the deletion already.
    INCARNATIONS = "*INCARNATIONS*" # M1Object   0  -1 All type-identical instances of this class. This is a delete blocker.
    MYATTRIBUTES = "*MYATTRIBUTES*" # M2Attribute   0  -1 The attribute definitions of this class.
    ATTRIBUTES = "*ATTRIBUTES*" # M2Attribute   0  -1 The attribute definitions of this and all superclasses.
    MYSRCASSOCIATIONS = "*MYSRCASSOCIATIONS*" # M2Association   0  -1 Associations, which start at this class.
    SRCASSOCIATIONS = "*SRCASSOCIATIONS*" # M2Association   0  -1 Associations, which start at this class or any superclass.
    MYDSTASSOCIATIONS = "*MYDSTASSOCIATIONS*" # M2Association   0  -1 Associations, which end at this class. Must not include any-associations
    DSTASSOCIATIONS = "*DSTASSOCIATIONS*" # M2Association   0  -1 Associations, which end at this class or any super class. Should include any-associations
    MYPARENTCOMPOSITIONS = "*MYPARENTCOMPOSITIONS*" # M2Composition   0  -1 Composition where I am the parent
    PARENTCOMPOSITIONS = "*PARENTCOMPOSITIONS*" # M2Composition   0  -1 Compositions where I am the parent, including the ones from superclasses.
    MYCHILDCOMPOSITIONS = "*MYCHILDCOMPOSITIONS*" # M2Composition   0  -1 Compositions where I am the child. 
    CHILDCOMPOSITIONS = "*CHILDCOMPOSITIONS*" # M2Composition   0  -1 Compositions where I am the child, including the ones from superclasses.
    MYSPECIALIZATIONS = "*MYSPECIALIZATIONS*" # M2Inheritance   0  -1 All inheritances where this class is the inheriting class.
    SPECIALIZATIONS = "*SPECIALIZATIONS*" # M2Inheritance   0  -1 All inheritances where this class or any superclass is the inheriting class.
    MYGENERALIZATIONS = "*MYGENERALIZATIONS*" # M2Inheritance   0  -1 All inheritances where this class is the superclass.
    GENERALIZATIONS = "*GENERALIZATIONS*" # M2Inheritance   0  -1 All inheritances where this class or any subclass is the superclass.
    
class M2ATTRIBUTE:
    def __init__(self):
        raise NotImplementedError()
    NAME = "*NAME" #   STR   1   1 The name of the attribute.
    CLASS = "*CLASS" # M2Class   1   1 The class this attribute definition is assigned to.
    PRIMTYPE = "*PRIMTYPE" #   STR   1   1 The primitive type, i.e. *BOL,*U32, *U64, ...., *STR, *ENU
    MUL = "*MUL" #   I64   1   1 The maximum number of values that can be stored in this attribute.
    UNIT = "*UNIT" #   STR   0   1 A unit string for that shall be valid for all values stored in this attribute.
    ENUM = "*ENUM" # M2Enum   0   1 An M2Enum, if PrimType is *NEU, else NON.
    INCARNATIONS = "*INCARNATIONS*" # M1Attribute   0  -1 All instances of this attribute definition in any M1 model.
    
class M2ASSOCIATION:
    def __init__(self):
        raise NotImplementedError()
    SRCNAME = "*SRCNAME" #   STR   1   1 The source name, i.e. the name used to navigate from the destination object to the source object(s).
    SRCCLASS = "*SRCCLASS" # M2Class   1   1 The source class.
    SRCMUL = "*SRCMUL" #   I64   1   1 The maximum number of elements on the source side of the association.
    DSTNAME = "*DSTNAME" #   STR   1   1 The destination name, i.e. the name used to navigate from the source object to the destination object(s).
    DSTCLASS = "*DSTCLASS" # M2Class   0   1 The destination class.
    DSTMUL = "*DSTMUL" #   I64   1   1 The maximum number of elements on the destination side of the association.
    ANYDST = "*ANYDST" #   BOL   1   1 If true, any elements of any type can be at the source. If true, dstClass must be NON.
    INCARNATIONS = "*INCARNATIONS*" # M1Association   0  -1 All instances of this association definition in any M1 model.
    
class M2COMPOSITION:
    def __init__(self):
        raise NotImplementedError()
    NAME = "*NAME" #   STR   1   1 The name of the composition.
    PARENTCLASS = "*PARENTCLASS" # M2Class   1   1 The parent class.
    CHILDCLASS = "*CHILDCLASS" # M2Class   0   1 The child class.
    MULCHILD = "*MULCHILD" #   I64   1   1 The maximum number of children in on the child side of this composition.
    ANYCHILD = "*ANYCHILD" #   BOL   1   1 If true, any elements of any type can be at the child. If true, childClass must be NON.
    INCARNATIONS = "*INCARNATIONS*" # M1Composition   0  -1 All instances of this composition
    
class M2INHERITANCE:
    def __init__(self):
        raise NotImplementedError()
    SUBCLASS = "*SUBCLASS" # M2Class   1   1 The inheriting class.
    SUPERCLASS = "*SUPERCLASS" # M2Class   1   1 The superclass.
    M1ATTRIBUTESBYINHERITANCE = "*M1ATTRIBUTESBYINHERITANCE*" # M1Attribute   0  -1 Delete blocker for inheritances. It is not valid to delete an inheritance if instances of the subclass exist.
    M1ASSOCIATIONSBYINHERITANCE = "*M1ASSOCIATIONSBYINHERITANCE*" # M1Association   0  -1 Delete blocker for inheritances. It is not valid to delete an inheritance if instances of the subclass exist.
    M1COMPOSITIONSBYINHERITANCE = "*M1COMPOSITIONSBYINHERITANCE*" # M1Composition   0  -1 Delete blocker for inheritances. It is not valid to delete an inheritance if instances of the subclass exist.
    
'''
M1 LAYER
The user model layer.
'''
class M1MODEL:
    def __init__(self):
        raise NotImplementedError()
    M2MODEL = "*M2MODEL" # M2Model   1   1 The M2 model instantiated by this M1 model.
    NAME = "*NAME" #   STR   1   1 The name of the M1 model.
    MODELROOT = "*MODELROOT" # M1Object   1   1 (optionally) The object which that is the root of the M1 model. Usually, all other M1 objects should be (indirect) children of this M1 object, i.e. it spans the model tree.
    OBJECTS = "*OBJECTS*" # M1Object   0  -1 All M1 objects in this model.
    
class M1OBJECT:
    def __init__(self):
        raise NotImplementedError()
    M2CLASS = "*M2CLASS" # M2Class   1   1 The M2 class instantiated.
    MODEL = "*MODEL" # M1Model   1   1 The M1 model this object belongs to.
    NAME = "*NAME" #   STR   1   1 The name of the M1 object.
    ATTRIBUTES = "*ATTRIBUTES*" # M1Attribute   0  -1 All attribute instances. One instance is one value of any attribute definition of the class.
    SRCASSOCIATIONS = "*SRCASSOCIATIONS*" # M1Association   0  -1 All associations instances where I am the source. One instance is a 1-to-1 link.
    DSTASSOCIATIONS = "*DSTASSOCIATIONS*" # M1Association   0  -1 All associations instances where I am the destination. One instance is a 1-to-1 link.
    PARENTCOMPOSITIONS = "*PARENTCOMPOSITIONS*" # M1Composition   0  -1 All composition instances where I am the parent. One instance is a 1-to-1 containment.
    CHILDCOMPOSITION = "*CHILDCOMPOSITION" # M1Composition   0   1 The composition instance where I am the child . It is a 1-to-1 containment.
    FEATUREVALUES = "*FVAL" #   VAL   0   1 
    FEATUREINSTANCES = "*FINS" # M1Feature   0   1 
    
class M1FEATURE:
    def __init__(self):
        raise NotImplementedError()
    
class M1ATTRIBUTE:
    def __init__(self):
        raise NotImplementedError()
    M2ATTRIBUTE = "*M2ATTRIBUTE" # M2Attribute   1   1 The M2 attribute this is an instance of.
    OBJECT = "*OBJECT" # M1Object   1   1 The M1 object this instance belongs to.
    VALUE = "*VALUE" #   PRM   1   1 The value of this instance. There is exactly one value per instance.
    POS = "*POS" #   U64   1   1 The position of the value in multi-value attributes, e.g. M2 attribute mul > 1.
    
class M1ASSOCIATION:
    def __init__(self):
        raise NotImplementedError()
    M2ASSOCIATION = "*M2ASSOCIATION" # M2Association   1   1 The M2 association this is an instance of.
    SRC = "*SRC" # M1Object   1   1 The M1 object being at the source end of this association.
    SRCPOS = "*SRCPOS" #   U64   1   1 The position of the src in multi-element associations, e.g. M2 association srcMul > 1.
    DST = "*DST" # MxElement   1   1 The M1 object being at the destination end of this association. Type must be MXELEMENT to allow ASSOCIATIONs to meta elements.
    DSTPOS = "*DSTPOS" #   U64   1   1 The position of the dst in multi-element associations, e.g. M2 association dstMul > 1.
    
class M1COMPOSITION:
    def __init__(self):
        raise NotImplementedError()
    M2COMPOSITION = "*M2COMPOSITION" # M2Composition   1   1 The M2 composition this is an instance of.
    PARENT = "*PARENT" # M1Object   1   1 The M1 object being at the parent of this composition.
    CHILD = "*CHILD" # MxElement   1   1 The M1 object being at the child ff this composition. Type must be MXELEMENT to allow ASSOCIATIONs to meta elements.
    POS = "*POS" #   U64   1   1 This is the position of the child. The parent position is always 1.
    

def IsConcept(name : str):
    return name.startswith(CONCEPT_PREFIX)

def IsMultivalueFeature(featureName : str):
    return featureName.endswith(FEATURE_MULTIVALUE_POSTFIX)

def NormalizeFeatureName(featureName : str):
    n = len(featureName)
    #start = 1 if featureName.startswith(GENERIC_FEATURE_PREFIX) or featureName.startswith(FEATURE_READONLY_PREFIX) else 0
    end = n-1 if featureName.endswith(FEATURE_MULTIVALUE_POSTFIX) else n
    return featureName[:end]

def GetConceptKeyString(conceptName:str)->str:
    '''Extracts the first and unique chars of a concept or concept feature name and returns it
    '''
    paddedName = conceptName+"    "
    return paddedName[CONCEPT_PREFIX_LEN:(CONCEPT_PREFIX_LEN+CONCEPT_UNIQUE_LEN)]
